from random import choice

from busy.command import QueueCommand


class DropCommand(QueueCommand):
    """Move items to the end of the todo collection of a queue"""

    name = 'drop'
    default_filter = [1]

    @QueueCommand.wrap
    def execute(self):
        collection = self.app.storage.get_collection(self.queue)
        if self.selection:
            lolist, hilist = self.collection.split(self.selection)
            self.collection.data = hilist + lolist
        self.set_next_item_status()


class PopCommand(QueueCommand):
    """Move items to the beginning of the collection"""

    name = 'pop'
    default_filter = ['-']

    @QueueCommand.wrap
    def execute(self):
        collection = self.app.storage.get_collection(self.queue)
        if self.selection:
            hilist, lolist = self.collection.split(self.selection)
            self.collection.data = hilist + lolist
        self.set_next_item_status()


class PickCommand(QueueCommand):
    """Move a random item to the beginning of the collection"""

    name = 'pick'
    default_filter = ['1-']

    @QueueCommand.wrap
    def execute(self):
        if self.selection:
            index = choice(self.selection)
            item = self.collection[index]
            hilist, lolist = self.collection.split([index])
            self.collection.data = hilist + lolist
        self.set_next_item_status()





from wizlib.parser import WizParser

from busy.command import CollectionCommand
from busy.util.checklist import Checklist


class PrintCommand(CollectionCommand):
    """Generate a Checklist PDF"""

    full: bool = False
    default_filter = ["1-"]
    name = "print"

    @classmethod
    def add_args(cls, parser: WizParser):
        super().add_args(parser)
        parser.add_argument('--long', '-l', action='store_true')

    @CollectionCommand.wrap
    def execute(self):  # pragma: nocover
        if self.long:
            tasks = [i.description for i in self.selected_items]
        else:
            tasks = [i.base for i in self.selected_items]
        queue = self.queue.capitalize()
        state = self.collection_state.capitalize()
        filter = (": "+",".join(self.filter)) \
            if (self.filter != self.default_filter) else ""
        title = f"{queue} ({state}{filter})"
        checklist = Checklist()
        checklist.generate(title, tasks)



import re
from busy.command import CollectionCommand
from busy.error import BusyError
from busy.model.item import Item

from wizlib.parser import WizParser

FIELDS = [r'^val:[a-z].*$', 'base', 'url']


class ViewCommand(CollectionCommand):
    """Output items using specified fields. Designed to replace base, describe,
    simple, etc. Defaults to the top item. Outputs crude comma separation."""

    name = 'view'
    fields: str = 'base'

    @classmethod
    def add_args(cls, parser: WizParser):
        super().add_args(parser)
        parser.add_argument('--fields', '-f', default='base')

    @CollectionCommand.wrap
    def execute(self):
        fields = self.fields.split(',')
        unknown_fields = [f for f in fields if
                          not any(re.match(p, f) for p in FIELDS)]
        if any(unknown_fields):
            raise BusyError(f"Unknown field(s) {','.join(unknown_fields)}")

        def row(item: Item):
            result = ''
            for field in fields:
                if field.startswith('val:'):
                    key = field[4]
                    val = item.data_value(key)
                else:
                    val = getattr(item, field)
                result += ',' if result else ''
                result += str(val) if val is not None else ''
            return result

        return self.output_items(row)



from datetime import datetime
from importlib import import_module

from wizlib.command import WizCommand
from wizlib.parser import WizParser
from wizlib.ui import Choice, Chooser
from wizlib.command import CommandCancellation

from busy.model.collection import Collection


class BusyCommand(WizCommand):

    default = 'simple'

    # TODO: Move to wizlib
    @staticmethod
    def add_yes_arg(parser: WizParser):
        parser.add_argument('--yes', '-y', action='store_true', default=None)

    # TODO: Move to wizlib
    def confirm(self, verb, *other_actions):
        """Ensure that a command is confirmed by the user"""
        if self.provided('yes'):
            return self.yes
        else:
            def cancel():
                raise CommandCancellation('Cancelled')
            chooser = Chooser(f"{verb}?", 'OK', [
                Choice('OK', '\n', True),
                Choice('cancel', 'c', cancel)
            ])
            for action in other_actions:
                name = action.name if hasattr(action, 'name') else 'other'
                key = action.key if hasattr(action, 'key') else 'o'
                chooser.add_choice(name, key, action)
            choice = self.app.ui.get_option(chooser)
            if type(choice) is bool:
                self.yes = choice
            return choice


class QueueCommand(BusyCommand):
    """Base for commands that work on the default collection of one queue"""

    queue: str = 'tasks'
    collection_state: str = 'todo'
    filter: list = None
    default_filter = [1]
    named_filters: list = []
    selection_optional = False

    @property
    def collection(self):
        """Return the collection object being queried, usually todo"""
        if not hasattr(self, '_collection'):
            self._collection = self.app.storage.get_collection(
                self.queue, self.collection_state)
        return self._collection

    @property
    def complete_filter(self):
        """Simple filters plus named filters"""
        return self.filter + self.named_filters

    @property
    def selection(self):
        """Indices of objects within the query collection that match the
        filter"""
        if not hasattr(self, '_selection'):
            self._selection = self.collection.select(*self.complete_filter)
        return self._selection

    @property
    def selected_items(self):
        """Items in the selection"""
        return [self.collection[i] for i in self.selection]

    @property
    def selected_items_list(self):
        """Simple text list of selected items"""
        return '\n'.join([i.listable for i in self.selected_items])

    def time_value(self, minutes):
        """Estimated value of effort, usable for billable hours, rounded to one
        decimal place"""
        if not hasattr(self, '_multiplier'):
            self._multiplier = self.app.config.get('busy-multiplier') or 1.0
        return self._multiplier * minutes / 60.0

    def summarize(self, items: list = None):
        if items is None:
            items = self.selected_items
        if len(items) == 1:
            result = "1 item"
        elif len(items) > 1:
            result = str(len(items)) + " items"
        else:
            return "nothing"
        elapsed = sum(i.elapsed for i in items)
        if elapsed:
            result += f" ({self.time_value(elapsed):.1f})"
        return result

    @classmethod
    def add_args(cls, parser: WizParser):
        super().add_args(parser)
        parser.add_argument('--queue', '-q', default='tasks', nargs='?')
        parser.add_argument('filter', nargs="*")

    def handle_vals(self):
        """Apply default filter"""
        super().handle_vals()
        self.update_current_task_time()
        if not self.provided('filter'):
            self.filter = self.default_filter

    def update_current_task_time(self):
        if self.queue == 'tasks':
            todos = self.app.storage.get_collection('tasks', 'todo')
            if len(todos) > 0:
                todos[0].update_time()
                todos.changed = True

    def start_current_task_timer(self):
        if self.queue == 'tasks':
            todos = self.app.storage.get_collection('tasks', 'todo')
            if len(todos) > 0:
                todos[0].start_timer()
                todos.changed = True

    @BusyCommand.wrap
    def execute(self, method, *args, **kwargs):
        """Handle timing and save operation"""
        result = None
        if self.selection_optional or self.selection:
            self.update_current_task_time()
            result = method(self, *args, **kwargs)
            self.start_current_task_timer()
            self.app.storage.save()
        return result

    def set_next_item_status(self):
        """Sets the status to the top item in the todo collection of this
        command's queue for convenient output"""
        todos = self.app.storage.get_collection(self.queue, 'todo')
        self.status = todos[0].simple if len(todos) else None


class CollectionCommand(QueueCommand):
    """Base for commands that work on a user-specified collection"""

    @classmethod
    def add_args(cls, parser: WizParser):
        super().add_args(parser)
        states = Collection.family_attrs('state')
        parser.add_argument(
            '--state', '-s', action='store', default='todo',
            dest='collection_state', choices=states)

    def output_items(self, func, with_index=False):
        """Return some attribute of all the items in the collection"""
        if with_index:
            return '\n'.join(func(self.collection[i], i)
                             for i in self.selection)
        else:
            return '\n'.join(func(i) for i in self.selected_items)


class NullCommand(QueueCommand):
    name = 'null'

    @QueueCommand.wrap
    def execute(self):
        pass


class IntegrationCommand:

    def execute_integration(self):
        module = import_module('busy_' + self.integration)
        integration = module.Main()
        func = getattr(integration, self.name)
        result = func(self)
        return result



import os
import re
from dataclasses import dataclass
from io import IOBase
from pathlib import Path

from busy.model.collection import Collection
from busy.storage import Storage

STATES = Collection.family_attrs('state')
RE_QUEUE = re.compile(f"([a-zA-Z0-9\\-]+)\\.({'|'.join(STATES)})\\.psv")


class FileStorage(Storage):

    def __init__(self, path: str = ''):
        if path:
            self.directory = Path(path) if isinstance(path, str) else path
        else:
            self.directory = Path.home() / '.busy'
        self.directory.mkdir(parents=True, exist_ok=True)
        assert isinstance(self.directory, Path) and self.directory.is_dir()
        self.cache = {}

    def filepath(self, queue: str, state: str) -> str:
        return self.directory / f"{queue}.{state}.psv"

    def get_collection(self, queue: str, state: str = 'todo'):
        id = (queue, state)
        if id not in self.cache:
            collection = Collection.family_member('state', state)()
            path = self.filepath(queue, state)
            if path.is_file():
                with open(path) as file:
                    collection.read_items(file)
            collection.changed = False
            self.cache[id] = collection
        return self.cache[id]

    def queue_exists(self, queue: str):
        collections = [self.get_collection(queue, s) for s in STATES]
        return any(len(c) for c in collections)

    @property
    def queue_names(self):
        """Return names of queues. Cache nothing."""
        result = set()
        for path in self.directory.iterdir():
            if path.is_file() and (match := RE_QUEUE.match(path.name)):
                queue = match.groups()[0]
                if (queue not in result) and path.stat().st_size:
                    result.add(queue)
        return result

    def save(self):
        """Save any changes and clear the cache"""
        while self.cache:
            key, collection = self.cache.popitem()
            if collection.changed:
                path = self.filepath(*key)
                with open(path, 'w') as file:
                    collection.write_items(file)




from wizlib.parser import WizParser

from busy.command import CollectionCommand


class DeleteCommand(CollectionCommand):

    yes: bool = None
    name = 'delete'
    key = 'd'
    default_filter = [1]

    @classmethod
    def add_args(cls, parser: WizParser):
        super().add_args(parser)
        cls.add_yes_arg(parser)

    def handle_vals(self):
        super().handle_vals()
        if self.selection:
            items = self.selected_items
            self.app.ui.send('\n'.join([str(i) for i in items]))
            self.confirm(f"Delete {self.summarize()}")

    # Assume the indices have been already set, before confirmation.

    @CollectionCommand.wrap
    def execute(self):
        deleted = self.collection.delete(self.selection)
        self.set_next_item_status()



from wizlib.parser import WizParser
from wizlib.command import CommandCancellation
from wizlib.ui.shell_ui import Emphasis

from busy.command import QueueCommand
from busy.error import BusyError
from busy.util import date_util


class Deferable(QueueCommand):

    @property
    def plan_date(self):
        return date_util.relative_date(self.when) \
            if self.when else None

    def ask_when(self):
        while True:
            self.when = self.app.ui.get_text(
                "When: ", [], (self.when or "tomorrow"))
            if self.check_when():
                break
        self.yes = True

    ask_when.name = 'other'
    ask_when.key = 'o'

    def check_when(self):
        if self.plan_date:
            return True
        else:
            self.app.ui.send(
                f"Invalid time description '{self.when}'", Emphasis.ERROR)

    def get_default_when(self):
        items = self.selected_items
        if not self.provided('when'):
            # Timing not provided so start with the default
            repeats = set(i.repeat_text for i in items if i.repeat_text)
            if len(repeats) > 1:
                raise BusyError(
                    'Items have different repeat values')
            if repeats:
                self.when = next(iter(repeats))
                if not self.check_when():
                    self.ask_when()


class DeferCommand(Deferable):

    when: str = ""
    yes: bool = None
    name = 'defer'

    @classmethod
    def add_args(cls, parser: WizParser):
        super().add_args(parser)
        parser.add_argument('--when', '-w')
        cls.add_yes_arg(parser)

    def handle_vals(self):
        super().handle_vals()
        if self.selection:
            self.app.ui.send(self.selected_items_list)
            self.get_default_when()
            if self.provided('when'):
                if not self.check_when():
                    self.ask_when()
            if not self.provided('yes'):
                if not self.when:
                    self.ask_when()
                if not self.provided('yes'):
                    intro = f"Defer {self.summarize()}"
                    intro += f" to {self.plan_date}"
                    self.confirm(intro, self.ask_when)

    @QueueCommand.wrap
    def execute(self):
        if not self.plan_date:
            raise BusyError('Invalid date')
        plans = self.app.storage.get_collection(self.queue, 'plan')
        items = self.collection.delete(self.selection)
        for item in items:
            item.plan(self.plan_date)
        plans += items
        self.set_next_item_status()



from busy.command import CollectionCommand
from busy.model.collection import Collection


class TagsCommand(CollectionCommand):

    name = 'tags'

    @CollectionCommand.wrap
    def execute(self):
        tags = set()
        for item in self.collection:
            tags |= item.tags
        return '\n'.join(sorted(tags))



from busy.command import BusyCommand, CollectionCommand, IntegrationCommand

from wizlib.parser import WizParser


class GetCommand(CollectionCommand, IntegrationCommand):
    """Gets a value from an integration"""

    name = 'get'

    @classmethod
    def add_args(cls, parser: WizParser):
        parser.add_argument('integration')
        super().add_args(parser)

    @BusyCommand.wrap
    def execute(self):
        return self.execute_integration()



from sys import version_info

MESSAGE = "This application requires Python version %i.%i.%i or higher"


def confirm_python_version(version):
    if version_info < version:
        raise RuntimeError(MESSAGE % version)





from busy.command import CollectionCommand
from busy.util.edit import edit_items


class EditorCommandBase(CollectionCommand):

    selection_optional = True

    @CollectionCommand.wrap
    def execute(self):
        command = self.app.config.get('editor') or 'emacs'
        edited = edit_items(self.collection,
                            self.selection, command)
        self.set_next_item_status()


class EditOneItemCommand(EditorCommandBase):
    """Edit items; default to just one"""

    name = "edit"


class EditManyCommand(EditorCommandBase):
    """Edit items; default to all"""

    name = 'manage'
    default_filter = ["1-"]



# Had an idea for a pluggable data storage interface. For now FileStorage is
# the only implementation.

from busy.model.collection import Collection


class Storage:
    """Abstract base class for Storage classes"""

    def get_collection(self, queue: str, state: str = 'todo') -> Collection:
        pass

    def save(self):
        pass



# Handy test case helpers, in the main library for plugin developers

from datetime import date, datetime
from unittest.mock import Mock, patch
from wizlib.test_case import WizLibTestCase


class BusyTestCase(WizLibTestCase):

    @staticmethod
    def mock_storage(o, p, d=None):
        def gc(q, s='todo'):
            if s == 'todo':
                return o
            elif s == 'plan':
                return p
            elif s == 'done':
                return d
        s = Mock()
        s.get_collection.side_effect = gc
        return s

    @staticmethod
    def patchtime(*t):
        return patch('busy.util.date_util.now',
                     lambda: datetime(*t))

    @staticmethod
    def patchday(*d):
        return patch('busy.util.date_util.today',
                     lambda: date(*d))



from wizlib.parser import WizParser

from busy.command import QueueCommand
from busy.model.item import Item


class AddCommand(QueueCommand):

    # description: str = ""
    name = 'add'
    selection_optional = True
    pop: bool = False

    @classmethod
    def add_args(cls, parser: WizParser):
        # Special case, no filter argument
        parser.add_argument('--queue', '-q', default='tasks', nargs='?')
        parser.add_argument('--pop', '-p', action='store_true', default=None)
        parser.add_argument('description', default="", nargs='?')

    def handle_vals(self):
        super().handle_vals()
        if not self.provided('description'):
            self.description = self.app.ui.get_text('Item: ')
            # edited = self.ui.edit_items(self.collection, self.selection)

    @QueueCommand.wrap
    def execute(self):
        if self.description:
            item = Item(self.description)
            if self.pop:
                self.collection.insert(0, item)
            else:
                self.collection.append(item)
        self.set_next_item_status()

    # @CollectionCommand.wrap
    # def execute(self):
    #     if not self.selection:
    #         self.status = "Edited nothing"
    #     else:
    #         edited = self.ui.edit_items(self.collection, self.selection)
    #         self.status = f"Edited {self.count(edited)}"
